//*******************************************************
//* CSTARTUP.C - startup C code module.					*
//* Copyright 2005 Everett Kaser						*
//* All rights reserved									*
//*******************************************************

#include	"cstartup.h"
#include	"game.h"

//********************************************************************************************************
//********************************************************************************************************
//********************************************************************************************************
// The MAIN.KA C runtime startup code calls (currently) 17 different routines as various
// events take place during the execution of the program.  These start with _KBeginStartup
// which is called VERY early in the startup of the KINT interpreter.  Nothing should be
// done at this point in terms of screen output or window creation, as the KINT interpreter
// hasn't even created its OWN "system" OS window yet.  This just gives the application a
// chance to do some early setup/initialization BEFORE window creation.  After KINT has finished 
// creating its system window and is about to drop into its normal event-handling loop, then it 
// calls _KEndStartup, at which point the KASM/KC code can start doing whatever it wants to in 
// terms of creating, setting up, initiatilizing, etc, the rest of its code.  _KBeginStartup calls 
// 'main()', which MUST exist in the applications code, and the primary job for main() is to HOOK 
// whatever events it wants to know about via the KHook() function, which enters the hook function 
// address in the KHookTab[] and returns the previous value that was in the table for that hook.  
// See the ON_ #defines in KWINLIB.H for indexes into the table.  If the application wants to ever
// be called again after main(), then it MUST hook SOME event (and usually ON_END_STARTUP).
// It makes no sense for main() to hook ON_BEGIN_STARTUP, since main() effectively IS OnBeginStartup,
// and ON_BEGIN_STARTUP only happens ONCE in the execution of the application, but it's #define is
// included in KWINLIB.H for the sake of completeness.
//********************************************************************************************************
//********************************************************************************************************
//********************************************************************************************************

extern	BOOL main();

static BOOL (*KHookTab[])() = {
	   0, 0, 0, 0, 0,
   &main, 0, 0, 0, 0,
	   0, 0, 0, 0, 0,
	   0, 0, 0, 0
};

//=============================================
// Install a new hook function and return the
// address of the old one.
BOOL (*KHook(long index, BOOL (*func)() ))() {

	BOOL (*oldfunc)();

	oldfunc = KHookTab[index];
	KHookTab[index] = func;
	return oldfunc;
}

//=============================================
void _KBeginStartup() {

	(*KHookTab[ON_BEGIN_STARTUP])();	// call application's main() function
}

//=============================================
void _KEndStartup() {

	if( KHookTab[ON_END_STARTUP] ) (*KHookTab[ON_END_STARTUP])();
}


//=============================================
int _KQueryQuit() {

	if( KHookTab[ON_QUERY_QUIT] ) return (*KHookTab[ON_QUERY_QUIT])();
	return 1;
}

//=============================================
// We're goin' down!!!  Swab the deck!!!
void _KShutDown() {

	if( KHookTab[ON_SHUT_DOWN] ) (*KHookTab[ON_SHUT_DOWN])();

}

//=============================================
// If an application hooks this event, it should return 0 to disable
// further processing of the event, otherwise if non-0 is returned
// then a WM_PARENTSIZED message is sent to the main window.
void _KWindowSize(long w, long h, long maxrestore) {

	if( KHookTab[ON_WINDOW_SIZE] ) if( !(*KHookTab[ON_WINDOW_SIZE])(w, h) ) return;
}

//=============================================
void _KWindowMove(int x, int y) {

	if( KHookTab[ON_WINDOW_MOVE] ) if( !(*KHookTab[ON_WINDOW_MOVE])(x, y) ) return;
}

//=============================================
void _KScreenChange() {

	if( KHookTab[ON_SCREEN_CHANGE] ) if( !(*KHookTab[ON_SCREEN_CHANGE])() ) return;
}

//=============================================
void _KWindowHidden(int minimized) {

	if( KHookTab[ON_WINDOW_HIDDEN] ) if( !(*KHookTab[ON_WINDOW_HIDDEN])(minimized) ) return;
}

//=============================================
// Mouse pointer has moved.  (x,y) is its new location and flags are button status:
//		bit 0: LBUTT is down
//		bit 1: RBUTT is down
//		bit 2: SHIFT is down
//		bit 3: CNTRL is down
//		bit 4: MBUTT is down
void _KPointerMove(int x, int y, int flags) {

	if( KHookTab[ON_POINTER_MOVE] ) if( !(*KHookTab[ON_POINTER_MOVE])(x, y, flags) ) return;
}

//=============================================
void _KAlphaKeyDown(int keycode) {

	if( KHookTab[ON_ALPHA_KEY_DOWN] ) if( !(*KHookTab[ON_ALPHA_KEY_DOWN])(keycode) ) return;
}

//=============================================
void _KAlphaKeyUp(int keycode) {

	if( KHookTab[ON_ALPHA_KEY_UP] ) if( !(*KHookTab[ON_ALPHA_KEY_UP])(keycode) ) return;
}

//=============================================
void _KKeyDown(DWORD keycode, DWORD repbutt, DWORD loc) {

	if( KHookTab[ON_KEY_DOWN] ) if( !(*KHookTab[ON_KEY_DOWN])(keycode, repbutt, loc) ) return;
}

//=============================================
void _KKeyUp(DWORD keycode, DWORD repbutt, DWORD loc) {

	if( KHookTab[ON_KEY_UP] ) if( !(*KHookTab[ON_KEY_UP])(keycode, repbutt, loc) ) return;
}

//=============================================
void _KChar(BYTE c) {

	if( KHookTab[ON_CHAR] )	if( !(*KHookTab[ON_CHAR])(c) ) return;
}

//=============================================
// The timer that was created by the application has gone off.
void _KTimer() {

	if( KHookTab[ON_TIMER] ) if( !(*KHookTab[ON_TIMER])(TimeMilli()) ) return;
}

//=============================================
// The machine is idle.  Return
//		0 if no need for further calls right now
//		1 if need to be called again soon
//
// NOTE: this feature can be problematic at best.  You should avoid
// using OnIdle if at all possible (always return 0), and use the OnTimer instead.
//
int _KIdle() {

	if( KHookTab[ON_IDLE] ) return (*KHookTab[ON_IDLE])();
	return 0;
}

//=============================================
// The music file that was playing has finished.
void _KMusicDone() {

	if( KHookTab[ON_MUSIC_DONE] ) if( !(*KHookTab[ON_MUSIC_DONE])() ) return;
}

//=============================================
// 'str' is ptr to string to set into program window's title bar
void SetProgramTitle(BYTE *str) {

	_asm(
		ldi r95,=0		@
		ldm r15,[str]	@
		SysSetPgmTitle
	);
}

//*******************************************************************************
//*******************************************************************************
//*******************************************************************************
//*
//* BITMAP FUNCTIONS
//*
//*******************************************************************************
//*******************************************************************************
//*******************************************************************************

//=============================================
HANDLE CreateBM(long w, long h) {

	return _asm(
		ldm r39,[w]		@
		ldm r38,[h]		@
		SysCreBM		@
		ldr r16,r15
	);
}

//=============================================
void DelBM(HANDLE hBM) {

	_asm(
		ldm r15,[hBM]	@
		tst r15			@
		je	@f			@
		SysDelBM		@
@:
	);
}

//=============================================
// Returns HANDLE or NULL if failed.
HANDLE LoadBM(BYTE *fname, short dir) {

	return _asm(
		ldm r15,[fname]	@
		ldi r95,=0		@
		ldm r94,[dir]	@
		SysLoadBM		@
		ldr r16,r15
	);
}

//=============================================
// Returns HANDLE or NULL if failed.
HANDLE LoadLossyBM(BYTE *fname, short dir) {

	return _asm(
		ldm r15,[fname]	@
		ldi r95,=0		@
		ldm r94,[dir]	@
		SysLoadLossyBM	@
		ldr r16,r15
	);
}

//=============================================
// Load a masked bitmap.  This supports two "transparent colors" (ie, the 'masked' colors)
// which don't get drawn.  You have to specify the (x,y) locations of those two pixels
// (which may be the SAME pixel if you only want one transparency color).
// Returns HANDLE or NULL if failed.
HANDLE LoadMBM(BYTE *fname, short dir, int transx1, int transy1, int transx2, int transy2) {

	return _asm(
		ldm r15,[fname]	@
		ldi r95,=0		@
		ldm r94,[dir]	@
		ldm r39,[transx1]@
		ldm r38,[transy1]@
		ldm r37,[transx2]@
		ldm r36,[transy2]@
		SysLoadMBM		@
		ldr r16,r15
	);
}

//=============================================
// Writes the (unmasked) bitmap to the file
//
void SaveBM(HANDLE hBM, BYTE *fname, short dir) {

	_asm(
		ldm r14,[fname]	@
		ldi r95,=0		@
		ldm r94,[dir]	@
		ldm r15,[hBM]	@
		SysSaveBM
	);
}

//=============================================
// Draw a masked or unmasked bitmap, normal or inverted.
// See BM_ defines in KWINLIB.H for values to pass as 'method'
// It's assumed that if you use BM_MASKED or BM_MASKED_INVERT
// that the srcBM IS a masked-bitmap.  The system COULD detect
// whether the bitmap has a mask or not and select BM_NOMASK
// or BM_MASKED automatically, but there are times when you
// WANT to draw BM_NOMASK with a masked srcBM.  Believe it.
//
void DrawBM(HANDLE dstBM, long dstX, long dstY,
			HANDLE srcBM, long srcX, long srcY, long srcW, long srcH, long method) {

	if( method & BM_MASKED ) {
		_asm(
			ldm r15,[dstBM]	@
			ldm r14,[srcBM]	@
			ldm r39,[dstX]	@
			ldm r38,[dstY]	@
			ldm r37,[srcX]	@
			ldm r36,[srcY]	@
			ldm r35,[srcW]	@
			ldm r34,[srcH]	@
			ldm r33,[method]@
			tst r33			@
			jod	DBMinvert1	@
			SysDrawMBM		@
			jmp @f			@
		DBMinvert1:			@
			SysInvDrawMBM	@
		@:
		);
	} else {
		_asm(
			ldm r15,[dstBM]	@
			ldm r14,[srcBM]	@
			ldm r39,[dstX]	@
			ldm r38,[dstY]	@
			ldm r37,[srcX]	@
			ldm r36,[srcY]	@
			ldm r35,[srcW]	@
			ldm r34,[srcH]	@
			ldm r33,[method]@
			tst r33			@
			jod	DBMinvert2	@
			SysDrawBM		@
			jmp @f			@
		DBMinvert2:			@
			SysInvDrawBM	@
		@:
		);
	}
}

//=============================================
// Draw a stretched masked or unmasked bitmap.  If unmasked,
// it can be stretched by pixel removal or duplication (BM_STRETCH_PIXEL)
// or by pixel blending (BM_STRETCH_SMOOTH).  If the srcBM is
// a masked bitmap, then smooth stretching is unavailable and
// pixel removal/duplication is always used (BM_STRETCH_MASKED).
// See BM_ defines in KWINLIB.H for values to pass as 'method.'
// It's assumed that if you use BM_STRETCH_MASKED that the 
// srcBM IS a masked-bitmap.
//
void StretchBM(HANDLE dstBM, long dstX, long dstY, long dstW, long dstH,
			HANDLE srcBM, long srcX, long srcY, long srcW, long srcH, long method) {

	if( method == BM_STRETCH_MASKED ) {
		_asm(
			ldm r15,[dstBM]	@
			ldm r14,[srcBM]	@
			ldm r39,[dstX]	@
			ldm r38,[dstY]	@
			ldm r37,[srcX]	@
			ldm r36,[srcY]	@
			ldm r35,[srcW]	@
			ldm r34,[srcH]	@
			ldm r33,[dstW]	@
			ldm r32,[dstH]	@
			SysStretchMBM
		);
	} else {
		_asm(
			ldm r15,[dstBM]	@
			ldm r14,[srcBM]	@
			ldm r39,[dstX]	@
			ldm r38,[dstY]	@
			ldm r37,[srcX]	@
			ldm r36,[srcY]	@
			ldm r35,[srcW]	@
			ldm r34,[srcH]	@
			ldm r33,[method]@
			tst r33			@
			jev	@f			@
			neg r35			@
		@:					@
			ldm r33,[dstW]	@
			ldm r32,[dstH]	@
			SysStretchBM
		);
	}
}

//=============================================
// Converts a NON-masked bitmap into a MASKED bitmap.
// 'trans1' and 'trans2' specify the two possible
// transparency colors (which may be the same value
// if you only have one transparency color).
//
void MaskBM(HANDLE hBM, DWORD trans1, DWORD trans2) {

	_asm(
		ldm r15,[hBM]	@
		ldm r14,[trans1]@
		ldm r13,[trans2]@
		SysMaskBM
	);
}

//=============================================
// Sets the 'orientation' (rotated/flipped) of a bitmap.
// For 'orient' see BMO_ #defines in KWINLIB.H
//
void OrientBM(HANDLE hBM, DWORD orient) {

	_asm(
		ldm r15,[hBM]	@
		ldm r39,[orient]@
		SysOrientBM
	);
}

//=============================================
// Returns (in *pBMI) the width, height, and orientation of the bitmap.
void GetBMinfo(HANDLE hBM, BMINFO *pBMI) {

	_asm(
		ldm r15,[hBM]	@
		SysSizeBM		@
		ldm r31,[pBMI]	@
		lde r15,r39		@
		pu0 r15,r31		@
		lde r15,r38		@
		pu0 r15,r31		@
		pu0 r37,r31
	);
}

//=============================================
// Make a variety of "colorization" changes to a bitmap region (x,y) by (w,h).
// If hue<0 then no hue changes are made, otherwise all pixels in the region are changed to that hue.
// If sat<0 then no changes are made.  If sat==0 then grayscale and hue is ignored.  Otherwise, all pixels use that saturation.
// 'lite' ranges from -255 to +255.  0 causes no ligthness (brightness) changes.  <0 makes pixels darker, >0 makes pixels brighter.
// 'contrast' ranges from -255 to +255.  0 == no change, <0 causes less contrast, >0 causes more contrast.
void ColorizeBM(HANDLE hBM, long x, long y, long w, long h, long hue, long sat, long lite, long contrast) {

	_asm(
		ldm r15,[hBM]	@
		ldm r39,[x]		@
		ldm r38,[y]		@
		ldm r37,[w]		@
		ldm r36,[h]		@
		ldm r35,[hue]	@
		ldm r34,[sat]	@
		ldm r33,[lite]	@
		ldm r32,[contrast]@
		SysColorize
	);
}

//=============================================
void ClipBM(HANDLE hBM, RECT r) {

	_asm(
		ldm r15,[hBM]	@
		ldm r39,[r.x1]	@
		ldm r38,[r.y1]	@
		ldm r37,[r.x2]	@
		ldm r36,[r.y2]	@
		sub r37,r39		@
		sub r36,r38		@
		SysClip
	);
}

//=============================================
void ClipBMXY(HANDLE hBM, long x1, long y1, long x2, long y2) {

	_asm(
		ldm r15,[hBM]	@
		ldm r39,[x1]	@
		ldm r38,[y1]	@
		ldm r37,[x2]	@
		ldm r36,[y2]	@
		sub r37,r39		@
		sub r36,r38		@
		SysClip
	);
}

//=============================================
RECT GetClipBM(HANDLE hBM) {

	RECT r;

	_asm(
		ldm r15,[hBM]	@
		SysGetClip		@
		ldm [r.x1],r39	@
		ldm [r.y1],r38	@
		add r37,r39		@
		ldm [r.x2],r37	@
		add r36,r38		@
		ldm [r.y2],r36	@
	);
	return r;
}


//*******************************************************************************
//*******************************************************************************
//*******************************************************************************
//*
//* TEXT/FONT FUNCTIONS
//*
//*******************************************************************************
//*******************************************************************************
//*******************************************************************************

// SEE ALSO: BeginText and EndText MACROS in KWINLIB.H

void DrawText(HANDLE hBM, long x, long y, BYTE *str) {

	_asm(
		ldm r15,[hBM]	@
		ldm r14,[str]	@
		ldi r95,=0		@
		ldm r39,[x]		@
		ldm r38,[y]		@
		SysText
	);
}

//=============================================
// Gets a font (if possible) that will allow the 'str' to
// be drawn within a w/h rectangle. If charW is non-NULL, it
// must point to a 512-byte (256-word) buffer which will be
// filled with the character widths (pixel-widths) of each of
// the 256 characters in the returned font.  If fontH and/or fontAH
// are non-NULL, then the HEIGHT of the font is stored in fontH and
// the ascending height of the font is stored in fontAH.
HANDLE FitFont(BYTE *str, long w, long h, long style, BYTE *charW, long *fontH, long *fontAH) {

	return _asm(
		ldm r13,[charW]	@
		ldi r94,=0		@
		ldm r14,[str]	@
		ldi r95,=0		@
		ldm r39,[w]		@
		ldm r38,[h]		@
		ldm r37,[style]	@
		SysGetFont		@
		ldm r15,[fontAH]@
		tst r15			@
		je	@f			@
		lde r16,r37		@
		ldm [r15],r16	@
	@:					@
		ldm r15,[fontH]	@
		tst r15			@
		je	@f			@
		lde r16,r38		@
		ldm [r15],r16	@
	@:					@
		lde r16,r39
	);
}

//=============================================
void RelFont(HANDLE hF) {

	_asm(
		ldm r39,[hF]	@
		SysRelFont
	);
}

//=============================================
// If backclr = -1 then the background (non-text pixels)
// are not drawn, otherwise they're filled with backclr.
void SetTextStyle(HANDLE hF, DWORD textclr, DWORD backclr) {

	_asm(
		ldm r39,[hF]		@
		ldm r13,[textclr]	@
		ldm r14,[backclr]	@
		SysTextStyle
	);
}

//=============================================
// returns WIDTH of string if drawn with current font.
// If h or ah are not NULL, then the HEIGHT and/or ASCENDING HEIGHT
// of the font are stored in those locations.
long GetTextSize(void *str, long *h, long *ah) {

	return _asm(
		ldm r14,[str]	@
		ldi r95,=0		@
		SysGetTextSize	@
		ldm r15,[h]		@
		tst r15			@
		je	@f			@
		lde r16,r38		@
		ldm [r15],r16	@
	@:					@
		ldm r15,[ah]	@
		tst r15			@
		je	@f			@
		lde r16,r37		@
		ldm [r15],r16	@
	@:					@
		lde r16,r39
	);
}

//*******************************************************************************
//*******************************************************************************
//*******************************************************************************
//*
//* GENERAL DRAWING FUNCTIONS
//*
//*******************************************************************************
//*******************************************************************************
//*******************************************************************************

void DrawPixel(HANDLE hBM, POINT p, DWORD color) {

	_asm(
		ldm r15,[hBM]	@
		ldm r14,[color]	@
		ldm r39,[p.x]	@
		ldm r38,[p.y]	@
		SysSetPixel
	);
}

void DrawPixelXY(HANDLE hBM, long x, long y, DWORD color) {

	_asm(
		ldm r15,[hBM]	@
		ldm r14,[color]	@
		ldm r39,[x]		@
		ldm r38,[y]		@
		SysSetPixel
	);
}

//=============================================
DWORD GetPixel(HANDLE hBM, POINT p) {

	return _asm(
		ldm r15,[hBM]	@
		ldm r39,[p.x]	@
		ldm r38,[p.y]	@
		SysGetPixel		@
		ldr r16,r14
	);
}

//=============================================
DWORD GetPixelXY(HANDLE hBM, long x, long y) {

	return _asm(
		ldm r15,[hBM]	@
		ldm r39,[x]		@
		ldm r38,[y]		@
		SysGetPixel		@
		ldr r16,r14
	);
}

//=============================================
// Set line style (bit pattern) used to draw lines and rectangle edges
// Returns previous line style.
DWORD LineStyle(DWORD ls) {

	return _asm(
		ldm r12,[ls]	@
		SysLineStyle	@
		ldr r16,r12
	);
}

//=============================================
// Draw lines.  '*p' is a pointer to an array of POINTs containing x/y coordinate
// pairs.  The first line starts at the first x/y point in the POINTs array and 
// goes to the second POINT.  Then a line is drawn from there (the second x/y point) 
// to the third, then from there to the fourth (do you see the sequence yet?)  
// The lines are all drawn in the color 'color'.
//
void DrawLines(HANDLE hBM, POINT *p, long numpoints, DWORD color) {

	_asm(
		ldm r15,[hBM]	@
		ldm r14,[p]		@
		ldi r95,=0		@
		ldm r13,[color]	@
		ldm r39,[numpoints]@
		SysLines
	);
}

//=============================================
// Draw a solid filled, outlined rectangle.  'fill' or 'edge' can be CLR_NONE (-1),
// otherwise they are RGB values in the 0x00BBGGRR format.
void Rectangle(HANDLE hBM, RECT r, DWORD fill, DWORD edge) {

	_asm(
		ldm r15,[hBM]	@
		ldm r14,[fill]	@
		ldm r13,[edge]	@
		ldm r39,[r.x1]	@
		ldm r38,[r.y1]	@
		ldm r37,[r.x2]	@
		sub r37,r39		@
		ldm r36,[r.y2]	@
		sub r36,r38		@
		SysRect
	);
}

//=============================================
// Draw a solid filled, outlined rectangle.  'fill' or 'edge' can be CLR_NONE (-1),
// otherwise they are RGB values in the 0x00BBGGRR format.
void RectangleXY(HANDLE hBM, long x1, long y1, long x2, long y2, DWORD fill, DWORD edge) {

	_asm(
		ldm r15,[hBM]	@
		ldm r14,[fill]	@
		ldm r13,[edge]	@
		ldm r39,[x1]	@
		ldm r38,[y1]	@
		ldm r37,[x2]	@
		ldm r36,[y2]	@
		sub r37,r39		@
		sub r36,r38		@
		SysRect
	);
}

//=============================================
// Draw a gradient filled rectangle.  Saturation varies left (minsat) to right (maxsat).
// Lightness varies bottom (minlite) to top (maxlite).  'hue' is same for every pixel.
void GradientRectangle(HANDLE hBM, RECT r, BYTE hue, BYTE minsat, BYTE maxsat, BYTE minlite, BYTE maxlite) {

	_asm(
		ldm r15,[hBM]	@
		ldi r14,=-2		@
		ldm r39,[r.x1]	@
		ldm r38,[r.y1]	@
		ldm r37,[r.x2]	@
		ldm r36,[r.y2]	@
		sub r37,r39		@
		sub r36,r38		@
		ldm r90,[hue]	@
		ldm r91,[minsat]@
		ldm r92,[maxsat]@
		ldm r93,[minlite]@
		ldm r94,[maxlite]@
		SysRect
	);
}

//=============================================
// Draw a solid filled, outlined oval.  'fill' or 'edge' can be CLR_NONE (-1),
// otherwise they are RGB values in the 0x00BBGGRR format.
void Oval(HANDLE hBM, RECT r, DWORD fill, DWORD edge) {

	_asm(
		ldm r15,[hBM]	@
		ldm r14,[fill]	@
		ldm r13,[edge]	@
		ldm r39,[r.x1]	@
		ldm r38,[r.y1]	@
		ldm r37,[r.x2]	@
		ldm r36,[r.y2]	@
		sub r37,r39		@
		sub r36,r38		@
		SysOval
	);
}

//=============================================
// Draw a solid filled, outlined oval.  'fill' or 'edge' can be CLR_NONE (-1),
// otherwise they are RGB values in the 0x00BBGGRR format.
void OvalXY(HANDLE hBM, long x1, long y1, long x2, long y2, DWORD fill, DWORD edge) {

	_asm(
		ldm r15,[hBM]	@
		ldm r14,[fill]	@
		ldm r13,[edge]	@
		ldm r39,[x1]	@
		ldm r38,[y1]	@
		ldm r37,[x2]	@
		ldm r36,[y2]	@
		sub r37,r39		@
		sub r36,r38		@
		SysOval
	);
}

//=============================================================
// Returns a millisecond counter that has no relationship to
// actual time, but can be used for measuring relative elapsed
// time intervals.  Every 1000 counts == 1 second.
DWORD TimeMilli() {

	return _asm(
		SysTimeMilli	@
		ldr r16,r15
	);
}

